import pandas as pd
from pandas.testing import assert_frame_equal

from test_common import CLEAN_NAME_DATA, DIRTY_COLUMN_NAMES, CLEAN_COLUMN_NAMES


def make_test_dfs():
    df = pd.DataFrame(CLEAN_NAME_DATA).T
    expected_df = df.copy()
    df.columns = DIRTY_COLUMN_NAMES
    expected_df.columns = CLEAN_COLUMN_NAMES
    return df, expected_df


def test_accessor_clean_names():
    df, expected_df = make_test_dfs()
    assert_frame_equal(df.bpw.clean_names, expected_df)


def test_accessor_samp():
    df = pd.DataFrame(CLEAN_NAME_DATA).T
    df.columns = CLEAN_COLUMN_NAMES
    df_samp = df.bpw.samp(2)
    assert_frame_equal(df.loc[df_samp.columns.tolist(), :].T, df_samp)



"""Module for the :class:`pandas.Series` accessor class."""

import functools

from pandas import Series
from pandas.api.extensions import register_series_accessor

from bpw_pde import ACCESSOR_NAME
from bpw_pde.common import clean_name, md5, unicode_normalize, DEFAULT_UNICODE_NORMALIZATION_FORM


@register_series_accessor(ACCESSOR_NAME)
class SeriesAccessor:
    """Accessor class for :class:`pandas.Series`."""

    def __init__(self, series: Series) -> None:
        self._series = series

    @property
    def clean_name(self) -> Series:
        """
        Applies :meth:`bpw_pde.common.clean_name` to the name of the captured :class:`pandas.Series`.

        :return: The captured :class:`pandas.Series` with a cleaned name.
        """

        self._series.name = clean_name(self._series.name)
        return self._series

    @property
    def hashify(self) -> Series:
        """
        Applies :meth:`bpw_pde.common.md5` to the captured :class:`pandas.Series`.

        :return: The captured :class:`pandas.Series` MD5 hashes.
        """

        return self._series.apply(md5)

    def normalize(self, form: str = DEFAULT_UNICODE_NORMALIZATION_FORM) -> Series:
        """
        Applies :meth:`bpw_pde.common.normalize` to the captured :class:`pandas.Series`.

        :param form: Unicode normalization form, defaults to `'NFKC'`.
        :return: The Unicode normalized captured :class:`pandas.Series`.
        """

        return self._series.apply(functools.partial(unicode_normalize, form=form))



import pandas as pd
from pandas.testing import assert_series_equal

from test_common import MD5_DATA, UNICODE_NORMALIZE_DATA


def make_clean_name_test_series():
    data = [1, 2, 3]
    series = pd.Series(data, name=' Foo!  ')
    expected_series = pd.Series(data, name='foo')
    return series, expected_series


def test_accessor_clean_names():
    series, expected_series = make_clean_name_test_series()
    assert_series_equal(series.bpw.clean_name, expected_series)


def test_accessor_md5():
    df = pd.DataFrame(MD5_DATA, columns=['text', 'expected'])
    assert_series_equal(df.text.bpw.hashify, df.expected, check_names=False)


def test_accessor_unicode_normalize():
    df = pd.DataFrame(UNICODE_NORMALIZE_DATA, columns=['text', 'expected'])
    assert_series_equal(df.text.bpw.normalize(), df.expected, check_names=False)



import pytest

from bpw_pde.common import clean_name, md5, unicode_normalize

DIRTY_COLUMN_NAMES = (' country      ', 'Phone Number', 'Dollar Amount (USD $)',)
CLEAN_COLUMN_NAMES = ('country', 'phone_number', 'dollar_amount_usd',)
CLEAN_NAME_DATA = [
    ['US', 'RU', 'TW', 'IS'],
    ['+01-301-555-8675', '+7 401 678 2741', '+886 4 5963 3071', '+354 276 5714'],
    ['12.34', '7.77', '123.45', '987.34'],
]
MD5_DATA = [
    ('Hello, I love you won\'t you tell me your name?', '41cfd1d06604733c5a40e89f71244537'),
    ('Riders on the storm ...', '206715483326fe890cd7151c0ff4e9aa'),
    ('Come on, baby, light my fire', 'feac31e22feec0a789bd98b0cebea8ad'),
]
UNICODE_NORMALIZE_DATA = [
    ('This should be a noop', 'This should be a noop'),
    ('Ⅷ', 'VIII'),
    ('ϔ', 'Ϋ'),
]


@pytest.mark.parametrize('name,expected', zip(DIRTY_COLUMN_NAMES, CLEAN_COLUMN_NAMES))
def test_clean_name(name: str, expected: str):
    assert clean_name(name) == expected


@pytest.mark.parametrize('text,expected', MD5_DATA)
def test_md5(text: str, expected: str):
    assert md5(text) == expected


@pytest.mark.parametrize('text,expected', UNICODE_NORMALIZE_DATA)
def test_unicode_normalize(text: str, expected: str):
    assert unicode_normalize(text) == expected



"""Module for the :class:`pandas.DataFrame` accessor class."""

import missingno
# noinspection PyUnresolvedReferences
import sidetable
from pandas import DataFrame
from pandas.api.extensions import register_dataframe_accessor

from bpw_pde import ACCESSOR_NAME
from bpw_pde.common import clean_name


class _MissignoAdapter:
    """
    Adapts the functional interface of the `missingno` package to the interface needed by :class:`DataframeAccessor`.

    Simply forwards the captured :class:`pandas.DataFrame` and any keyword arguments to the respective `missingno`
    function. This class is not intended to be used directly but through the :class:`DataframeAccessor`.
    """

    def __init__(self, df: DataFrame) -> None:
        self._df = df

    def matrix(self, **kwargs):
        return missingno.matrix(self._df, **kwargs)

    def bar(self, **kwargs):
        return missingno.bar(self._df, **kwargs)

    def heatmap(self, **kwargs):
        return missingno.heatmap(self._df, **kwargs)

    def dendrogram(self, **kwargs):
        return missingno.dendrogram(self._df, **kwargs)


@register_dataframe_accessor(ACCESSOR_NAME)
class DataframeAccessor:
    """Accessor class for :class:`pandas.DataFrame`."""

    def __init__(self, df: DataFrame) -> None:
        self._df = df

    @property
    def clean_names(self) -> DataFrame:
        """
        Applies :meth:`bpw_pde.common.clean_name` to all column names of the captured :class:`pandas.DataFrame`.

        :return: The captured :class:`pandas.DataFrame` with cleaned column names.
        """

        self._df.columns = [clean_name(column) for column in self._df.columns]
        return self._df

    def samp(self, n: int = 5) -> DataFrame:
        """
        Transposed random sample of the captured :class:`pandas.DataFrame`.

        It is common want to take a quick look at a couple of random samples of rows from a :class:`pandas.DataFrame`
        and have it all fit on screen without mucking with `pandas` display settings. This usually does a better job
        of that than just calling :meth:`pandas.DataFrame.sample` alone.

        :param n: The number of rows to sample, defaults to `5`.
        :return: A random sample of the captured :class:`pandas.DataFrame` rows transposed.
        """

        return self._df.sample(n).T

    @property
    def stb(self):
        """
        Alias to `sidetable` :class:`pandas.DataFrame` accessor `stb`.

        :return: The `stb` accessor on the captured :class:`pandas.DataFrame`.
        """

        return self._df.stb

    def freq(self, *args, **kwargs):
        """
        Alias to :meth:`sidetable.SideTableAccessor.freq`.

        :param args: ``args`` forwarded to :meth:`sidetable.SideTableAccessor.freq`.
        :param kwargs: ``kwargs`` forwarded to :meth:`sidetable.SideTableAccessor.freq`.
        :return: The result of calling :meth:`sidetable.SideTableAccessor.freq` on the captured :class:`pandas.DataFrame`.
        """

        return self._df.stb.freq(*args, **kwargs)

    @property
    def msno(self):
        """
        Adapted `missingno` interface to work as a nested accessor.

        :return: An adapter class that provides access to the functions in `missingno`.
        """

        return _MissignoAdapter(self._df)



"""
Module for the :class:`pandas.Index` accessor class.

Currently, this module has no useful functionality.
"""

from pandas import Index
from pandas.api.extensions import register_index_accessor

from bpw_pde import ACCESSOR_NAME


@register_index_accessor(ACCESSOR_NAME)
class IndexAccessor:
    """Accessor class for :class:`pandas.Index`."""

    def __init__(self, index: Index) -> None:
        self._index = index



"""Common functions used across the index, series, and dataframe modules."""

import hashlib
import string
import unicodedata

PUNCTUATION_TO_SPACE_TRANSLATOR = str.maketrans(string.punctuation, ' ' * len(string.punctuation))
DEFAULT_UNICODE_NORMALIZATION_FORM = 'NFKC'


def clean_name(name: str) -> str:
    """
    Applies lowercasing, whitespace tokenization, and rejoining tokens with underscore to ``name``.

    :param name: The name to be cleaned.
    :return: The cleaned name.
    """
    return '_'.join(name.lower().translate(PUNCTUATION_TO_SPACE_TRANSLATOR).split())


def md5(text: str) -> str:
    """
    Computes the MD5 hash of ``text``.

    :param text: The text to be hashed using the MD5 hashing algorithm.
    :return: The MD5 hash of the text.
    """
    return hashlib.md5(text.encode('utf-8')).hexdigest()


def unicode_normalize(text: str, form: str = DEFAULT_UNICODE_NORMALIZATION_FORM) -> str:
    """
    Applies Unicode normalization on ``text`` using the form ``form``.

    :param text: The text to be normalized.
    :param form: The Unicode normalization form to use for normalization.
    :return: The Unicode normalized text.
    """
    return unicodedata.normalize(form, text)



"""Top-level bpw_pde package."""

import logging
from logging import NullHandler

from tqdm.autonotebook import tqdm

__author__ = 'Bryan Patrick Wood'
__email__ = 'bpw1621@gmail.com'
__version__ = '0.0a2'

logging.getLogger(__name__).addHandler(NullHandler())
tqdm.pandas()

ACCESSOR_NAME = 'bpw'

from .dataframe import DataframeAccessor
from .index import IndexAccessor
from .series import SeriesAccessor
